<?php

namespace App\Command;

use ApiPlatform\Core\Api\IriConverterInterface;
use Ramsey\Uuid\Uuid;
use App\Entity\Core\Widget;
use Psr\Log\LoggerInterface;
use App\Entity\Core\ImportJob;
use App\Constants\AppConstants;
use App\Constants\ErrorConstants;
use App\Service\SettingsHelper;
use App\Entity\Core\SolutionPack;
use App\Entity\Core\ExportTemplate;
use App\Entity\Core\Tag;
use Doctrine\ORM\EntityManagerInterface;
use App\Service\SolutionPackUtilityService;
use Symfony\Component\Filesystem\Filesystem;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\ArrayInput;
use Symfony\Component\Console\Input\InputOption;
use App\Service\ConfigExportImport\ImportService;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\BufferedOutput;
use Symfony\Component\Console\Output\OutputInterface;
use App\Service\ConfigExportImport\ImportOptionsGenerator;
use App\Service\UtilityService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class SolutionPackHandlerCommand extends Command
{
    const SOLUTIONPACK_TYPE = 'solutionpack';
    const TMP_PATH = "/tmp/solutionpacks/";
    const DUPLICATEMESSAGE = "DUPLICATEMESSAGE";
    const SOLUTIONPACK = 'solutionPack';
    const DATA_ROUTING_KEY = 'key.api.notify';
    const DATA_EXCHANGE = "cyops.crudhub.datanotify";

    private $entityManager;
    private $containers;

    private $settingsHelper;
    private $importService;
    private $logger;
    private  $filesystem;
    private $iriConverter;
    private $importOptionsGenerator;
    private $solutionPackUtilityService;
    private $utilityService;

    public function __construct(
        ContainerInterface $containers,
        EntityManagerInterface $entityManager,
        ImportService $importService,
        SolutionPackUtilityService $solutionPackUtilityService,
        UtilityService $utilityService,
        ImportOptionsGenerator $importOptionsGenerator,
        SettingsHelper $settingsHelper,
        LoggerInterface $logger,
        IriConverterInterface $iriConverter
    ) {
        $this->containers = $containers;
        $this->entityManager = $entityManager;
        $this->importService = $importService;
        $this->solutionPackUtilityService = $solutionPackUtilityService;
        $this->utilityService = $utilityService;
        $this->settingsHelper = $settingsHelper;
        $this->importOptionsGenerator = $importOptionsGenerator;
        $this->logger = $logger;
        $this->iriConverter = $iriConverter;
        $this->filesystem = new Filesystem();
        parent::__construct();
    }

    protected function configure(): void
    {
        $this
            ->setName('app:solutionpack:import')
            ->setAliases(['fortisoar:solutionpack:import'])
            ->setDefinition([
                new InputOption(
                    'message',
                    'm',
                    InputOption::VALUE_OPTIONAL,
                    'Arguments for importing solution pack.'
                )
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        try {
            $stringifiedMessage =  base64_decode($input->getOption('message'));
            $message = json_decode($stringifiedMessage, true);
            foreach ($message as $value) {
                if (!isset($value['entityType'])) {
                    return 0;
                }
                $this->logger->info("Operation Received : " . $value['operation']);
                $this->logger->info("entityType Received : " . $value['entityType']);

                if ($this->solutionPackUtilityService->checkPrimaryNode()) {
                    switch ($value['entityType']) {
                        case 'Widget':
                            if ($value['operation'] == "Delete") {
                                $this->deleteWidget($value);
                            } else if ($value['operation'] == 'Install') {
                                $this->installWidget($value);
                            } else if ($value['operation'] == 'Update') {
                                $this->updateWidget($value);
                            }
                            break;
                        case 'connector':
                            if ($value['operation'] == 'Install Success' || $value['operation'] == 'Create' || $value['operation'] == 'Version Created') {
                                $this->installIntegration($value);
                            } else if ($value['operation']  == 'Uninstall Success') {
                                $this->uninstallIntegration($value);
                            } else if ($value['operation']  == 'Publish') {
                                $this->updateIntegration($value, true);
                            } else if ($value['operation']  == 'AddConfig') {
                                $this->integrationAddConfig($value);
                            } else if ($value['operation']  == 'DeleteConfig') {
                                $this->integrationDeleteConfig($value);
                            }
                            break;
                        case self::SOLUTIONPACK:
                            if ($value['operation'] == 'install' || $value['operation'] == 'update' || $value['operation'] == 'dependencyInstall') {
                                $this->installSolutionPack($value);
                            }
                            break;
                        default:
                            break;
                    }
                    return 0;
                } else {
                    $payload = array(
                        "message" => [$value],
                        "routing_key" => self::DATA_ROUTING_KEY,
                        "exchange" => self::DATA_EXCHANGE
                    );

                    $objectData = array(
                        "handlerType" => "message_handler",
                        "nodeIds" => "PRIMARY_NODE",
                        "payload" => $payload,
                        "commandType" => "publish"
                    );
                    $dasUri = $this->containers->getParameter('das.uri');
                    $dasClient = $this->containers->get("app.proxy.client.das");
                    try {
                        $dasRequests = new GuzzleRequest('post', $dasUri . '/clustercommand/', ["Content-type" => "application/json"], json_encode($objectData));
                        $dasResponse = $dasClient->send($dasRequests);
                        $resCode = $dasResponse->getStatusCode();
                        if ($resCode != 200) {
                            return 0;
                        }
                    } catch (\Exception $error) {
                        error_log('cyops-auth service request failed');
                        error_log($error->getMessage());
                        return 0;
                    }
                }
                return 0;
            }
        } catch (\Exception $e) {
            error_log($e->getMessage());
            return 0;
        }
    }

    private function validateMinimumVersion($solutionRecord, $minVersion)
    {
        if ($minVersion) {
            $minimalVersion = str_replace('.', '', $minVersion);
            $availableVersion =  str_replace('.', '', $solutionRecord->getVersion());
            if ($availableVersion >= $minimalVersion) {
                if ($solutionRecord->getImportJob()) {
                    $this->solutionPackUtilityService->updateImportJob($solutionRecord->getImportJob(), AppConstants::STATUS['IMPORT_COMPLETE'], null, null, null, 'completed', 'Import completed.');
                }
            } else {
                $latestAvailableVersion = $solutionRecord->getLatestAvailableVersion();
                if ($latestAvailableVersion && $availableVersion <  str_replace('.', '', $latestAvailableVersion)) {
                    $solutionRecord->setVersion($latestAvailableVersion);
                }
                if ($solutionRecord->getImportJob()) {
                    $this->entityManager->remove($solutionRecord->getImportJob());
                    $this->entityManager->flush();
                }
                $importJob = new ImportJob();
                $importJob->setUuid(Uuid::uuid4()->toString());
                $importJob->setStatus(AppConstants::STATUS['DRAFT']);
                $importJob->setType('SolutionPack Import');
                $this->entityManager->persist($importJob);
                $solutionRecord->setFile("");
                $solutionRecord->setImportJob($importJob);
                return true;
            }
        } else {
            if ($solutionRecord->getImportJob()) {
                $this->solutionPackUtilityService->updateImportJob($solutionRecord->getImportJob(), AppConstants::STATUS['IMPORT_COMPLETE'], null, null, null, 'completed', 'Import completed.');
            }
        }
        return false;
    }

    private function validateAndAppendImportedBy($solutionRecord, $message)
    {
        $existingImportBy = false;
        $importArray = [];
        $existingImport = $solutionRecord->getImportedBy();
        if ($existingImport) {
            foreach ($existingImport as $value) {
                if ($value['name'] == $message['importedBy'][0]['name']) {
                    $existingImportBy = true;
                }
            }
            if (!$existingImportBy) {
                array_push($existingImport, $message['importedBy'][0]);
                $importArray = $existingImport;
            }
        } else {
            $importArray =  $message['importedBy'];
        }
        $solutionRecord->setImportedBy($importArray);
    }

    private function installSolutionPack($message)
    {
        $this->entityManager = $this->containers->get('doctrine')->resetManager();
        $minVersion = $message['minVersion'];
        $solutionRecord = $this->solutionPackUtilityService->getRecord($message['name']);
        if (!$solutionRecord) {
            return false;
        }
        if (array_key_exists('previousDep', $message) && !is_null($message['previousDep'])) {
            $l_message = sprintf("Current solution pack => %s is dependent on %s", $message['name'], $message['previousDep']);
            $this->logger->info($l_message);
            $previousDep = $message['previousDep'];
            $depRecord = $this->solutionPackUtilityService->getRecord($previousDep);
            if (is_null($depRecord) || !$depRecord->getInstalled()) {
                if ($message['operation'] != 'dependencyInstall') {
                    $l_message = sprintf("Dependent solution pack => %s of %s is not present ", $message['previousDep'], $message['name']);
                    $this->logger->info($l_message);
                    $this->solutionPackUtilityService->updateImportJob($solutionRecord->getImportJob(), AppConstants::STATUS['ERROR'], null, null, ErrorConstants::DEPENDENCY_FAILED);
                    $this->solutionPackUtilityService->removeFileObj($solutionRecord, false, true, true);
                    $solutionRecord->setInstalled(false);
                    $solutionRecord->setStatus(AppConstants::STATUS['ERROR']);
                    $this->entityManager->persist($solutionRecord);
                    $this->entityManager->flush();
                }
                return false;
            }

        }
        $l_message = sprintf("Installing solution pack %s", $message['name']);
        $this->logger->info($l_message);
        $solutionRecord = $this->solutionPackUtilityService->getRecord($message['name']);
        if (!$solutionRecord) {
            return false;
        }
        //section to add the importedby on the entity which are going to be installed as part of dependency.
        if ($message['operation'] == 'dependencyInstall' && $message['importedBy']) {
            $this->validateAndAppendImportedBy($solutionRecord, $message);
        }
        if ($solutionRecord->getInstalled() && !$solutionRecord->getLocal() && $message['operation'] != "update") {
            $upgradeRequired = $this->validateMinimumVersion($solutionRecord, $minVersion);
            if (!$upgradeRequired) {
                return true;
            }
        }

        $importedBy = $this->getImportedBy($solutionRecord);
        $solutionRecord->setStatus("Installing");
        $this->entityManager->persist($solutionRecord);
        $this->entityManager->flush();
        if ($message['tokenType'] == 'JWT') {
            $token = $this->getRefreshToken($message['userId']);
        } else {
            $token['token'] = $message['userToken'];
        }
        $this->settingsHelper->setRefreshedToken($token['token'], $message['tokenType']);
        list($fileBasePath, $fileExtractedPath, $fileName) = $this->solutionPackUtilityService->loadFolderIntoTmp($solutionRecord);

        $input = new ArrayInput([]);
        $output = new BufferedOutput();
        $io = new SymfonyStyle($input, $output);
        $this->removeImportFolderMissingInInfo($fileExtractedPath);
        $this->generateOptions($solutionRecord, $fileBasePath, $fileExtractedPath, $fileName, $io, $message['tokenType']);
        $this->logger->info("options generated sucessfully");
        $importResult = $this->executeImportJob($solutionRecord->getImportJob()->getUuid(), $fileExtractedPath, $io, $importedBy, $message['tokenType']);
        if ($importResult['success']) {
            $this->generateExportTemplate($solutionRecord, $fileExtractedPath);
            $this->solutionPackUtilityService->updateSolutionPack($solutionRecord, "Completed", true);
            $this->filesystem->remove($fileBasePath);
            $this->solutionPackUtilityService->removeFileObj($solutionRecord, false);
        } else {
            $this->solutionPackUtilityService->removeFileObj($solutionRecord, false, true, true);
            $solutionRecord->setInstalled(false);
            $solutionRecord->setStatus(AppConstants::STATUS['ERROR']);
            $this->entityManager->persist($solutionRecord);
            $this->entityManager->flush();
            $this->filesystem->remove($fileBasePath);
            throw $importResult['error'];
        }
        return true;
    }

    private function getRefreshToken($userId)
    {
        $data = array(
            'uuid' => $userId,
            'isReport' => true
        );
        $dasUri = $this->containers->getParameter('das.uri');
        $keyFile = file_get_contents('/opt/cyops-api/src/keys/daspublic.key', true);
        $idKey = base64_encode($keyFile);
        $data['idKey'] = $idKey;
        $headers = array(
            "Content-type" => "application/json",
        );
        $dasClient = $this->containers->get("app.proxy.client.das");
        $dasRequests = new GuzzleRequest('POST', $dasUri . '/token?uuid=' . $data['uuid'], $headers, json_encode($data));
        try {
            $dasResponse = $dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            return json_decode($dasResponseData, true);
        } catch (BadRequestHttpException | \Exception $error) {
            $errVar = json_decode($error->getResponse()->getBody()->read(2048), true);
            error_log($errVar['Error']);
            throw new \Exception("Error in getting renewed token from auth service ");
        }
    }


    private function generateOptions(SolutionPack $solutioPackRecord, $fileBasePath, $fileExtractedPath, $fileName, $io, $tokenType = 'JWT')
    {
        try {
            $this->logger->info("generateOptions with  solutio pack record: " . $solutioPackRecord->getName());
            if ($solutioPackRecord->getImportJob()) {
                $importJobUuid = $solutioPackRecord->getImportJob()->getUuid();
            } else {
                $importJobUuid = $this->createImportJob($solutioPackRecord, $fileBasePath, $fileName);
            }
            $repository = $this->entityManager->getRepository(ImportJob::class);
            /** @var ImportJob $importJob */
            $importJob = $repository->find($importJobUuid);
            $this->solutionPackUtilityService->updateImportJob($importJob, null, 0, [], null, null, []);
            // Generate options for each key
            $this->solutionPackUtilityService->updateImportJob($importJob, AppConstants::STATUS['GENERATING_OPTIONS'], 10, null, null, 'starting', 'Generating Options...');
            $this->logger->info("Import Options Generator for importJobs :" . $importJobUuid);
            $options = $this->importOptionsGenerator->generateFromFolder($fileExtractedPath, $importJob, $io, $tokenType);
            $this->solutionPackUtilityService->updateImportJob($importJob, AppConstants::STATUS['OPTIONS_GENERATED'], 100, $options, null, null, 'Import options generated successfully');
            $this->logger->info("Import Options Generated :" . $options);
            return $options;
        } catch (\Exception $e) {
            $this->solutionPackUtilityService->removeFileObj($solutioPackRecord, false, true, true);
            $solutioPackRecord->setStatus(AppConstants::STATUS['ERROR']);
            $solutioPackRecord->setInstalled(false);
            $this->entityManager->persist($solutioPackRecord);
            $this->entityManager->flush();
            $this->filesystem->remove($fileBasePath);
            error_log("Failed to generate options");
            $message = sprintf('Error while generating options for solution pack : %s', $e->getMessage());
            $this->logger->error($message);
            throw $e;
        }
    }

    private function executeImportJob($importJobUuid = null, $fileExtractedPath = "", $io = null, $importedBy = [], $tokenType = 'JWT')
    {
        $this->logger->info("executeImportJob with  import Job Uuid: " . $importJobUuid);
        $repository = $this->entityManager->getRepository(ImportJob::class);
        /** @var ImportJob $importJob */
        $importJob = $repository->find($importJobUuid);
        if (is_null($importJob)) {
            $msg = "Import job does not exist - Uuid = " . $importJobUuid;
            return ['failure' => true, 'error' => $msg];
        }
        $cmd = '';
        try {
            $this->solutionPackUtilityService->updateImportJob($importJob, AppConstants::STATUS['IMPORTING'], 0, null, null, null, 'Beginning import');
            $configImportResult = $this->importService->importConfigurationFromFolder($fileExtractedPath, $importJob, $io, $tokenType, $importedBy);
            $this->logger->info("Executing Import Command: " . $cmd);
            if (!$configImportResult['success']) {
                throw new \Exception('Configuration import failed.');
            }
            $command = '/usr/bin/php /opt/cyops-api/bin/console app:cache:clear';
            exec($command, $execOutput, $returnVal);
            $cmd = 'sudo -u nginx /usr/bin/php /opt/cyops-api/bin/console app:cache:clear';
            $this->utilityService->broadcastToHA('app_cache_clear', $cmd);
            $this->solutionPackUtilityService->updateImportJob($importJob, AppConstants::STATUS['IMPORT_COMPLETE'], null, null, null, 'completed', 'Import completed.');
            return $configImportResult;
        } catch (\Exception $e) {
            $this->logger->error("Failed to generate options");
            $message = sprintf('Error while executing ImportJob for solution pack : %s', $e->getMessage());
            $this->logger->error($message);
            // Returning as throwing result in Solution Pack not getting set to false and UI continuous shows spinning
            return ['failure' => true, 'error' => $e];
        }
    }

    private function installIntegration($message)
    {
        try {
            $this->logger->info("installIntegration with  : " . $message['data']['name']);
            $repository = $this->entityManager->getRepository(SolutionPack::class);
            $solutionPack = $repository->findOneBy(array("name" => $message['data']['name']));

            // If Connector is incompatable delete call for connector wont be published.
            // So value of installed will be true in other case it will always be false
            if (is_null($solutionPack) || $solutionPack->getInstalled()) {
                $solutionPack = new SolutionPack();
                $solutionPack->setVersion('0.0.0');
                $solutionPack->setInfoContent(['help' => $message['data']['help'] ?? $message['data']['help_online']]);
            }
            $versionNumber = str_replace('.', '', $message['data']['version']);
            $availableVersionNumber =  str_replace('.', '', $solutionPack->getVersion());
            if ($availableVersionNumber > $versionNumber) {
                $solutionPack->setLatestAvailableVersion($solutionPack->getVersion());
            } else {
                $solutionPack->setLatestAvailableVersion(null);
            }
            $solutionPack->setName($message['data']['name']);
            $solutionPack->setVersion($message['data']['version']);
            $solutionPack->setDevelopment($message['data']['development']);
            $solutionPack->setInstalled($message['data']['installed']);
            $solutionPack->setType('connector');
            $solutionPack->setRecordId($message['data']['id']);
            if ($message['data']['rpm_installed']) {
                $solutionPack->setLocal(false);
                $solutionPack->setDevelopment(false);
            }
            if ($solutionPack->getLocal()) {
                $solutionPack->setPublisher($message['data']['publisher']);
            }
            $solutionPackInfoJson = $solutionPack->getInfoContent();
            $solutionPackInfoJson['help'] = $message['data']['help_online'] ?? $solutionPackInfoJson['help'];
            $solutionPack->setInfoContent($solutionPackInfoJson);
            // setCertified is not called since custom upload should not be certified .
            // $solutionPack->setInfoPath();
            // $solutionPack->setDependencies();
            $solutionPack->setIconLarge($message['data']['icon_large']);
            $solutionPack->setDescription($message['data']['description']);
            $solutionPack->setLabel($message['data']['label']);
            $solutionPack->setStatus("Installed");
            $tags = $this->solutionPackUtilityService->importTags($message['data']['tags']);
            foreach ($tags as $tag) {
                $solutionPack->addRecordTag($this->iriConverter->getItemFromIri('/api/3/tags/' . $tag, ['resource_class' => Tag::class, 'fetch_data' => false]));
            }
            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $this->logger->error("Exception while install integration: " . $e->getMessage());
        }
    }

    private function updateIntegration($message, $publish = false)
    {
        try {
            $this->logger->info("update integration with  : " . $message['data']['name']);
            $repository = $this->entityManager->getRepository(SolutionPack::class);
            if ($message['data']['workspace_connector']) {
                $solutionPack = $repository->findOneBy(array("name" => $message['data']['name'], "version" => $message['data']['version']));
            } else {
                $solutionPack = $repository->findOneBy(array("name" => $message['data']['name']));
            }
            if (is_null($solutionPack)) {
                $solutionPack = new SolutionPack();
                $solutionPack->setVersion('0.0.0');
                $solutionPack->setInfoContent(['help' => $message['data']['help'] ?? $message['data']['help_online']]);
            }
            $versionNumber = str_replace('.', '', $message['data']['version']);
            $availableVersionNumber =  str_replace('.', '', $solutionPack->getVersion());
            if ((int)$availableVersionNumber > (int)$versionNumber) {
                $solutionPack->setLatestAvailableVersion($solutionPack->getVersion());
            } else {
                $solutionPack->setLatestAvailableVersion(null);
            }
            $solutionPack->setName($message['data']['name']);
            $solutionPack->setVersion($message['data']['version']);
            $solutionPack->setDevelopment($message['data']['development']);
            $solutionPack->setInstalled($message['data']['installed']);
            $solutionPack->setType('connector');
            $solutionPack->setRecordId($message['data']['id']);
            $publish ? $solutionPack->setLocal(true) : $solutionPack->setLocal($message['data']['development']);
            if ($solutionPack->getLocal()) {
                $solutionPack->setPublisher($message['data']['publisher']);
            }
            $solutionPackInfoJson = $solutionPack->getInfoContent();
            $solutionPackInfoJson['help'] = $message['data']['help_online'] ?? $solutionPackInfoJson['help'];
            $solutionPack->setInfoContent($solutionPackInfoJson);
            // setCertified is not called since custom upload should not be certified .
            // $solutionPack->setInfoPath();
            // $solutionPack->setDependencies();
            //Add category handling
            $solutionPack->setIconLarge($message['data']['icon_large']);
            $solutionPack->setDescription($message['data']['description']);
            $solutionPack->setLabel($message['data']['label']);
            $tags = $this->solutionPackUtilityService->importTags($message['data']['tags']);
            foreach ($tags as $tag) {
                $solutionPack->addRecordTag($this->iriConverter->getItemFromIri('/api/3/tags/' . $tag, ['resource_class' => Tag::class, 'fetch_data' => false]));
            }

            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $this->logger->error("Exception while updateIntegration: " . $e->getMessage());
        }
    }

    private function uninstallIntegration($payload)
    {
        $repository = $this->entityManager->getRepository(SolutionPack::class);
        $this->logger->info("uninstallIntegration Name" . $payload['data']['name'] . " - Version" . $payload['data']['version']);
        $solutionPack = $repository->findOneBy(array("name" => $payload['data']['name'], "version" => $payload['data']['version']));
        try {
            if ($solutionPack && !$solutionPack->getLocal()) {
                $this->solutionPackUtilityService->removeFileObj($solutionPack, true, true, true);
                $solutionPack->setDraft(false);
                $solutionPack->setInstalled(false);
                $solutionPack->setRecordId(null);
                $solutionPack->setImportJob(null);
                $solutionPack->setFile(null);
                $solutionPack->setExportJob(null);
                $solutionPack->setStatus(null);
                $solutionPack->setTemplate(null);
                $this->entityManager->persist($solutionPack);
                $this->entityManager->flush();
            } else if ($solutionPack && $solutionPack->getLocal()) {
                $this->solutionPackUtilityService->removeFileObj($solutionPack, true, true, true);
                $this->entityManager->remove($solutionPack);
                $this->entityManager->flush();
            }
        } catch (\Exception $e) {
            $this->logger->error("Exception while integration un-install:" . $e->getMessage());
        }
    }

    private function deleteWidget($message)
    {
        try {
            $this->logger->info("deleteWidget with UUID : " . $message['data']['uuid']);
            $repository = $this->entityManager->getRepository(SolutionPack::class);
            $solutionPack = $repository->findOneBy(array("recordId" => $message['data']['uuid']));
            if (!is_null($solutionPack) && !$solutionPack->getLocal()) {
                $this->solutionPackUtilityService->removeFileObj($solutionPack, true, true, true);
                $solutionPack->setDraft(false);
                $solutionPack->setInstalled(false);
                $solutionPack->setRecordId(null);
                $solutionPack->setImportJob(null);
                $solutionPack->setExportJob(null);
                $solutionPack->setFile(null);
                $solutionPack->setStatus(null);
                $solutionPack->setTemplate(null);
                $this->entityManager->persist($solutionPack);
                $this->entityManager->flush();
            } else if ($solutionPack && $solutionPack->getLocal()) {
                $this->solutionPackUtilityService->removeFileObj($solutionPack, true, true, true);
                $this->entityManager->remove($solutionPack);
                $this->entityManager->flush();
            }
        } catch (\Exception $e) {
            $this->logger->error("Exception while deleteWidget: " . $e->getMessage());
        }
    }

    private function installWidget($message)
    {
        try {
            $this->logger->info("installWidget for: " . $message['data']['name'] . " - Version :" . $message['data']['version']);
            $repository = $this->entityManager->getRepository(SolutionPack::class);
            $solutionPack = $repository->findOneBy(array("name" => $message['data']['name']));
            if (is_null($solutionPack)) {
                $solutionPack = new SolutionPack();
            }
            $versionNumber = str_replace('.', '', $message['data']['version']);
            $availableVersionNumber =  str_replace('.', '', $solutionPack->getVersion());
            if ((int)$availableVersionNumber > (int)$versionNumber) {
                $solutionPack->setLatestAvailableVersion($solutionPack->getVersion());
            } else if ($availableVersionNumber == $versionNumber) {
                $solutionPack->setLatestAvailableVersion(null);
            }
            $solutionPack->setName($message['data']['name']);
            $solutionPack->setVersion($message['data']['version']);
            $solutionPack->setDraft($message['data']['draft']);
            $solutionPack->setDevelopment($message['data']['draft']);
            $solutionPack->setLocal($message['data']['local']);
            $solutionPack->setInstalled($message['data']['installed'] ?? false);
            $solutionPack->setType('widget');
            $solutionPack->setRecordId($message['data']['uuid']);
            $solutionPack->setPublisher($message['data']['metadata']['publisher']);
            // setCertified is not called since custom upload should not be certified .
            $solutionPack->setInfoContent(['metadata' => $message['data']['metadata'], "subTitle" => $message['data']['subTitle']]);
            $solutionPack->setIconLarge($message['data']['icon_large']);
            $solutionPack->setDescription($message['data']['subTitle'] ?? $message['data']['metadata']['description']);
            $message['data']['label'] = $message['data']['label'] ?? $message['data']['title'];
            $solutionPack->setLabel($message['data']['label']);
            $solutionPack->setPublishedDate($message['data']['publishedDate']);
            $solutionPack->setStatus("Installed");

            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $this->logger->error("Exception while installWidget: " . $e->getMessage());
        }
    }

    private function updateWidget($message)
    {
        $this->logger->info("updateWidget for: " . $message['data']['name'] . " - Version :" . $message['data']['version']);
        $repository = $this->entityManager->getRepository(SolutionPack::class);
        $solutionPack = $repository->findOneBy(array("name" => $message['data']['name'], "version" => $message['data']['version']));
        $repository = $this->entityManager->getRepository(Widget::class);
        $widget = $repository->findOneBy(array("name" => $message['data']['name']));
        if (is_null($widget)) {
            return false;
        }
        try {
            if (is_null($solutionPack)) {
                $solutionPack = new SolutionPack();
                $solutionPack->setName($widget->getName());
                $solutionPack->setVersion($widget->getVersion());
                $solutionPack->setType('widget');
                $solutionPack->setRecordId($widget->getUuid());
                $solutionPack->setStatus("Installed");
            }
            $versionNumber = str_replace('.', '', $message['data']['version']);
            $availableVersionNumber =  str_replace('.', '', $solutionPack->getVersion());
            if ((int)$availableVersionNumber > (int)$versionNumber) {
                $solutionPack->setLatestAvailableVersion($solutionPack->getVersion());
            } else if ($availableVersionNumber == $versionNumber) {
                $solutionPack->setLatestAvailableVersion(null);
            }
            $solutionPack->setDraft($widget->getDraft());
            $solutionPack->setDevelopment($widget->getDraft());
            $solutionPack->setInstalled($widget->getInstalled());
            // setCertified is not called since custom upload should not be certified .
            $solutionPack->setInfoContent(['metadata' => $widget->getMetadata(), "subTitle" => $widget->getSubTitle()]);
            $solutionPack->setDescription($widget->getSubTitle() ?? $widget->getMetadata()['description']);
            $solutionPack->setPublishedDate($widget->getPublishedDate());
            $solutionPack->setPublisher($widget->getMetadata()['publisher']);
            $widgetLabel = $widget->getLabel() ?? $widget->getTitle();
            $solutionPack->setLabel($widgetLabel);

            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $this->logger->error("Exception while integration configuration addition." . $e->getMessage());
        }
    }

    private function integrationAddConfig($message)
    {
        $this->logger->info("integrationAddConfig Name" . $message['data']['name'] . " - Version" . $message['data']['version']);
        $repository = $this->entityManager->getRepository(SolutionPack::class);
        $solutionPack = $repository->findOneBy(array("name" => $message['data']['name'], "version" => $message['data']['version']));
        if (is_null($solutionPack)) {
            return true;
        }
        try {
            $solutionPack->setConfigCount((int)$solutionPack->getConfigCount() + 1);
            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $this->logger->error("Exception while integration configuration addition:" . $e->getMessage());
        }
    }

    private function integrationDeleteConfig($message)
    {
        $this->logger->info("integrationDeleteConfig Name" . $message['data']['name'] . " - Version" . $message['data']['version']);
        $repository = $this->entityManager->getRepository(SolutionPack::class);
        $solutionPack = $repository->findOneBy(array("name" => $message['data']['name'], "version" => $message['data']['version']));
        if (is_null($solutionPack)) {
            return true;
        }
        try {
            if ((int)$solutionPack->getConfigCount() > 0) {
                $solutionPack->setConfigCount((int)$solutionPack->getConfigCount() - 1);
                $this->entityManager->persist($solutionPack);
                $this->entityManager->flush();
            }
        } catch (\Exception $e) {
            $this->logger->error("Exception while integration configuration removal:" . $e->getMessage());
        }
    }

    private function createImportJob($solutioPackRecord, $fileBasePath, $fileName)
    {
        $this->logger->info("createImportJob for solution pack with base path:" . $fileBasePath);
        try {
            $file = $this->solutionPackUtilityService->uploadFile($fileBasePath, $fileName);
            $importJob = new ImportJob();
            $importJob->setFile(AppConstants::FILE_IRI . $file->getUuid());
            $importJob->setUuid(Uuid::uuid4()->toString());
            $importJob->setStatus(AppConstants::STATUS['DRAFT']);
            $importJob->setType('SolutionPack Import');
            $this->entityManager->persist($importJob);
            $this->entityManager->flush();
            if ($solutioPackRecord) {
                //since the record is avilabe in DB only update the status.
                $solutioPackRecord->setFile(AppConstants::FILE_IRI . $file->getUuid());
                $solutioPackRecord->setImportJob($importJob);
                $this->entityManager->persist($solutioPackRecord);
                $this->entityManager->flush();
            }
            return  $importJob->getUuid();
        } catch (\Throwable $th) {
            $this->filesystem->remove($fileBasePath);
            $this->logger->error("Failed to create import job.");
            throw $th;
        }
    }

    private function getImportedBy($solutionPack)
    {
        return [['apiName' => $solutionPack->getName(), 'name' =>  $solutionPack->getLabel(), 'version' => $solutionPack->getVersion()]];
    }

    private function generateExportTemplate($solutionRecord, $fileExtractedPath)
    {
        $filteredOption = [];
        $this->logger->info("generateExportTemplate for solution pack Uuid:" . $solutionRecord->getUuid());
        $options = $solutionRecord->getImportJob()->getOptions();
        foreach ($options as $key => $value) {
            $filteredOption[$key] = $value['values'];
        }
        try {
            $result = $this->importService->exportOptionGenerationFromInputOptions($filteredOption, $fileExtractedPath);
            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $exportTemplate = new ExportTemplate();
            $exportTemplate->setOptions($result);
            $exportTemplate->setName("FortiSOAR-" . $solutionRecord->getName() . "-" . $solutionRecord->getVersion() . "-" . $timestamp->format('Y-m-d H:i:s'));
            $exportTemplate->setType("SolutionPack Export");
            $exportTemplate->setUuid(Uuid::uuid4()->toString());
            $this->entityManager->persist($exportTemplate);
            $this->entityManager->flush();
            $solutionRecord->setTemplate($exportTemplate);
            $this->entityManager->persist($solutionRecord);
            $this->entityManager->flush();
        } catch (\InvalidArgumentException | \Exception $e) {
            $this->logger->error("InvalidArgumentException at generateExportTemplate: " . $e->getMessage());
            return false;
        }
    }

    private function removeImportFolderMissingInInfo($infoJsonPath)
    {
        if (file_exists($infoJsonPath . 'info.json')) {
            $infoJsonContent = json_decode(file_get_contents($infoJsonPath . 'info.json'), true);
            //check for multicontent if required
            if ($infoJsonContent['type'] == self::SOLUTIONPACK_TYPE) {
                $sDir = scandir($infoJsonPath);
                if (isset($infoJsonContent['contents']['picklistNames'])) {
                    $infoJsonContent['contents']['picklists'] = $infoJsonContent['contents']['picklistNames'];
                }
                if (isset($infoJsonContent['contents']['recordSets'])) {
                    $infoJsonContent['contents']['records'] = $infoJsonContent['contents']['recordSets'];
                }
                $contentKeys = array_keys($infoJsonContent['contents']);
                foreach ($sDir as $value) {
                    if (!in_array($value, array(".", "..", 'images')) &&  is_dir($infoJsonPath . $value) && (!in_array($value, $contentKeys) || str_starts_with($value, "__"))) {
                        $this->filesystem->remove($infoJsonPath . $value);
                    }
                }
            }
        }
    }
}
