<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller\TAXII;

use App\Providers\QueryProvider;
use App\Security\Voter\FsrVoter;
use App\Constants\ErrorConstants;
use App\Entity\Query\SystemQuery;
use App\Entity\Custom\ThreatIntel;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use ApiPlatform\Core\DataProvider\CollectionDataProviderInterface;
use ApiPlatform\Core\Serializer\SerializerContextBuilderInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

class TaxiiApiController extends AbstractController
{

    const THREAT_FEED_MODULE_NAME = 'threat_intel_feeds';

    private $entityManager;
    private $serializer;
    private $fsrVoter;
    private $queryProvider;

    public function __construct(
        EntityManagerInterface $entityManager,
        SerializerInterface $serializer,
        SerializerContextBuilderInterface $serializerContextBuilder,
        CollectionDataProviderInterface $collectionDataProvider,
        FsrVoter $fsrVoter,
        QueryProvider $queryProvider,
        ContainerInterface $containers
    ) {
        $this->entityManager = $entityManager;
        $this->serializer = $serializer;
        $this->serializerContextBuilder = $serializerContextBuilder;
        $this->collectionDataProvider = $collectionDataProvider;
        $this->fsrVoter = $fsrVoter;
        $this->queryProvider = $queryProvider;
        $this->containers = $containers;
    }

    /**
     * @Route("api/taxii/1/taxii", name="app_api_root", methods={"GET"})
     */
    public function getTaxiiRootApi()
    {
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException();
        }

        $wfClient = $this->containers->get("app.proxy.client.workflow");
        $headers = array(
            "Content-type" => "application/json",
        );
        $workflowUri = $this->containers->getParameter('workflow.uri');
        $workflowRequests = new GuzzleRequest('get', $workflowUri . '/api/dynamic-variable/?format=json&name=Server_fqhn', $headers);
        $system_host = 'localhost';
        try {
            $workflowResponse = $wfClient->send($workflowRequests);
            $workflowResponseData = json_decode($workflowResponse->getBody()->getContents(), true);
            if ($workflowResponseData["hydra:totalItems"] > 0) {
                $system_host = $workflowResponseData["hydra:member"][0]["value"];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            error_log($error->getMessage());
        }

        $responseObject = [
            "api_roots" => [
                "https://$system_host/api/taxii/1/"
            ],
            "contact" => "fsr-integrations@fortinet.com",
            "default" => "https://$system_host/api/taxii/1/",
            "description" => "TAXII 2.1 Server",
            "title" => "FortiSOAR Taxii 2.1 Server"
        ];
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($responseObject));
        return $response;
    }

    /**
     * @Route("api/taxii/1/", name="app_api_root_info", methods={"GET"})
     */
    public function getTaxiiRootApiInfo()
    {
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException();
        }
        $responseObject = [
            "max_content_length" => 1000000,
            "title" => "TAXII feeds",
            "versions" => [
                "taxii-2.1"
            ]
        ];
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($responseObject));
        return $response;
    }

    /**
     * @Route("api/taxii/1/collections", name="app_api_feeds_collections", methods={"GET"})
     */
    public function getFeedsCollection()
    {
        $module = self::THREAT_FEED_MODULE_NAME;
        $systemSettingRepo = $this->entityManager->getRepository(SystemSettings::class);
        $taxiiSettingObj = $systemSettingRepo->findOneBy(['name' => "TAXII Server Details"]);
        if (!is_null($taxiiSettingObj) && !$taxiiSettingObj->getPublicValues()['status']['enabled']) {
            throw new BadRequestHttpException(ErrorConstants::TAXII_SERVER_DISABLED);
        }
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException("Access Denied");
        }
        $systemQueryRepo = $this->entityManager->getRepository(SystemQuery::class);
        $feedCollectionsFilters = $systemQueryRepo->createQueryBuilder('o')
            ->join('o.models', 'm')
            ->andWhere('m.type = :types')
            ->setParameter('types', $module)
            ->getQuery()->getResult();

        $collectionList = array();
        $collectionList['collections'] = array();
        if (!empty($feedCollectionsFilters)) {
            foreach ($feedCollectionsFilters as $filter) {
                $collectionList['collections'][] = [
                    "name" => $filter->getName(),
                    "id" => $filter->getUuid(),
                    "uuid" => $filter->getUuid()
                ];
            }
        }
        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($collectionList));
        return $response;
    }


    private function getRequestContext($request, $attributes)
    {
        $request->overrideGlobals();
        $context = $this->serializerContextBuilder->createFromRequest($request, true, $attributes);
        $context["relationships"] =  false;
        $context["ignore_authorization"] =  true;
        $context["ignore_field_authorization"] =  true;
        $context["ignore_relationship_authorization"] =  true;
        $context["displayNameOnly"] =  true;
        return $context;
    }

    // Support Feed Listing
    // Support Filter Feed by Feed ID, Collection Name
    // Support Filter Feed by combination of Feed ID and Collection Name
    // All Response return are in Taxii 2.1 format

    /**
     * @Route("api/taxii/1/collections/{collectionId}/{objects?}/{objectId?}", name="app_api_feeds_export", requirements={"objectId"=".+"}, methods={"GET"})
     */
    public function getFeeds(Request $request, $collectionId, $objects = 'objects', $objectId = null)
    {
        $systemSettingRepo = $this->entityManager->getRepository(SystemSettings::class);
        $taxiiSettingObj = $systemSettingRepo->findOneBy(['name' => "TAXII Server Details"]);
        if (!is_null($taxiiSettingObj) && !$taxiiSettingObj->getPublicValues()['status']['enabled']) {
            throw new BadRequestHttpException(ErrorConstants::TAXII_SERVER_DISABLED);
        }
        $module = self::THREAT_FEED_MODULE_NAME;
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
            throw new AccessDeniedException("Access Denied");
        }
        $resourceClass = ThreatIntel::class;
        $queryLimit = $request->get('$limit', 30);
        $queryPage = $request->get('$page', 1);
        $partial = $request->get('$partial', false);
        $attributes = [];
        $attributes['filters'] = ['$limit' => $queryLimit, '$page' => $queryPage, '$partial' => $partial];
        $attributes['resource_class'] = $resourceClass;
        $attributes['collection_operation_name'] = "get";
        $systemQueryRepo = $this->entityManager->getRepository(SystemQuery::class);
        $savedFilter = $systemQueryRepo->findOneBy(['uuid' => $collectionId]);
        if (is_null($savedFilter)) {
            throw new BadRequestHttpException(sprintf('Collection ID %s not found', $collectionId));
        }

        if (is_null($objects)) {
            $response = new Response();
            $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
            $collectionDetails = [
                "name" => $savedFilter->getName(),
                "id" => $savedFilter->getUuid(),
                "uuid" => $savedFilter->getUuid()
            ];

            return $response->setContent(json_encode($collectionDetails));
        }

        $filterObj = $savedFilter->getQuery();
        foreach ($filterObj['filters'] as $index => $filterValue) {
            if (is_array($filterValue['value'])) {
                $filterObj['filters'][$index]['value'] = [$filterValue['value'][0]['@id']];
            }
        }
        if (!is_null($objectId)) {
            $filterObj['filters'][$index + 1] = [
                "field" => "uuid",
                "operator" => "like",
                "type" => "primitive",
                "value" => '%' . $objectId . '%'
            ];
        }
        $this->queryProvider->setInitialQuery($filterObj);
        $context = $this->getRequestContext($request, $attributes);
        $feedByCollection = $this->collectionDataProvider->getCollection($resourceClass, $attributes['collection_operation_name'], $attributes);
        if (empty($feedByCollection)) {
            throw new NotFoundHttpException();
        }
        $normalized_data = $this->serializer->normalize($feedByCollection, 'jsonld', $context);
        unset($normalized_data['@context']);
        unset($normalized_data['@id']);
        unset($normalized_data['@type']);
        unset($normalized_data['hydra:totalItems']);
        if (is_array($normalized_data["hydra:view"]) && isset($normalized_data["hydra:view"]["hydra:next"])) {
            $normalized_data['more'] = true;
            $normalized_data['next'] = $queryPage + 1;
        }
        $normalized_data['objects'] = $normalized_data['hydra:member'];
        unset($normalized_data['hydra:member']);
        unset($normalized_data["hydra:view"]);

        $response = new Response();
        $response->headers->set('Content-Type', 'application/taxii+json;version=2.1');
        $response->setContent(json_encode($normalized_data));

        return $response;
    }
}
