<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\ModelMetadata\ModelMetadata;
use App\Entity\ViewTemplate\SystemViewTemplate;

/**
 * Class ViewConfig
 * @package App\Service\ConfigExportImport
 */
class ViewConfig extends PortingConfig
{
    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $typeKey, $tokenType)
    {
        $data = $allData[$typeKey];
        $options = [];
        $repository = $this->loadRepository(SystemViewTemplate::class);
        foreach ($data as $key => $value) {
            if (!$value) {
                continue;
            }
            if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                // This is a sequential array, used by legacy module view templates
                foreach ($value as $view) {
                    $existing = $repository->findOneBy(['uuid' => $view['uuid']]);
                    $individualOptions = [];
                    $uuid = $view['uuid'];
                    if (!$uuid) {
                        $uuid = implode('-', explode('/', $view['id']));
                    }
                    $individualOptions['uuid'] = $uuid;
                    if ($view['config'] && array_key_exists('title', $view['config'])) {
                        $individualOptions['name'] = $view['config']['title'];
                    }
                    $individualOptions['module'] = $key;
                    $individualOptions['exists'] = $existing ? true : false;
                    $individualOptions['include'] = true;

                    $options[] = $individualOptions;
                }
                continue;
            }
            $existing = $repository->findOneBy(['uuid' => $value['uuid']]);
            $individualOptions = [];
            $uuid = $value['uuid'];
            $pieces = explode('-', $value['uuid']);
            if (count($pieces) === 3) {
                $individualOptions['module'] = $pieces[1];
            }
            if ($value['config'] && array_key_exists('title', $value['config'])) {
                $individualOptions['name'] = $value['config']['title'];
            }
            if ($value['uuid'] == 'app') {
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    foreach ($existing->getConfig()['navigation'] as $existingItem) {
                        if ($existingItem['title'] == $item['title']) {
                            $value['config']['navigation'][$key]['exists'] = true;
                            $value['config']['navigation'][$key]['include'] = true;
                        }
                    }
                }
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    if (is_null($value['config']['navigation'][$key]['exists']) || $value['config']['navigation'][$key]['exists'] == false) {
                        $value['config']['navigation'][$key]['exists'] = false;
                        $value['config']['navigation'][$key]['include'] = true;
                    }
                }
            }
            $individualOptions['uuid'] = $uuid;
            $individualOptions['exists'] = $existing ? true : false;
            $individualOptions['include'] = true;
            $individualOptions['config'] = $value['config'];

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        foreach ($options as $option) {
            if (!$option['include']) {
                // Do not merge skipped items
                continue;
            }
            $option['appendNavigation']  =  $option['appendNavigation'] ?? true;
            $foundView = null;
            foreach ($data as $value) {
                if (!$value) {
                    continue;
                }
                if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                    // This is a sequential array, used by module view templates
                    foreach ($value as $view) {
                        if ($option['uuid'] == $view['uuid']) {
                            $foundView = $view;
                            break;
                        }
                    }
                    continue;
                }
                if ("app" == $value['uuid'] && $option['appendNavigation']) {
                    $fullURI = sprintf('%s/api/views/1/app', $this->container->getParameter('crudhub.uri'));
                    $result = json_decode($this->sendRequest($fullURI, 'GET', null, $tokenType), true);
                    $includeItems = $result["config"]["navigation"];
                    foreach ($option['config']['navigation'] as $key => $optionValue) {
                        if ($optionValue['include'] == true || $optionValue['include'] == "true") {
                            $itemExists = false;
                            unset($optionValue['include']);
                            unset($optionValue['exists']);
                            foreach ($includeItems as $key => $navitem) {
                                if ($navitem['title'] == $optionValue['title']) {
                                    $mergerdValue = array_merge($navitem, $optionValue);
                                    $includeItems[$key] = $mergerdValue;
                                    $itemExists = true;
                                    break;
                                }
                            }
                            if (!$itemExists) {
                                array_push($includeItems, $optionValue);
                            }
                        }
                    }
                    $result['config']['navigation'] = $includeItems;
                    $foundView = $result;
                } elseif ($option['uuid'] == $value['uuid']) {
                    $foundView = $value;
                    break;
                }
            }
            if ($option['exists'] && $foundView) {
                $toUpdate[] = $foundView;
            } else if ($foundView) {
                $toAdd[] = $foundView;
            }
        }
        $results = [];
        if (!empty($toAdd)) {
            $fullURI = sprintf('%s/api/3/', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $toAdd, $tokenType);
            $results[] = $result;
        }
        if (!empty($toUpdate)) {
            $fullURI = sprintf('%s/api/3/update/system_view_templates', $this->container->getParameter('crudhub.uri'));
            $body = [
                "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        foreach ($options as $value) {
            // if modules has details, export correlation settings also
            if (strpos($value, 'detail') !== false) {
                array_push($options, str_replace('detail', 'settings', $value));
            }
        }
        return $this->getResourceFromIRI($options, SystemViewTemplate::class);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modifiedOptions = [];
        $appendNavigation = [];
        $metaData = [];
        $key = null;
        foreach ($options as $value) {
            if (is_string($value)) {
                $modifiedOptions[] = $value;
            }
            if (is_array($value) && !empty($value)) {
                $modifiedOptions[] = $value['value'];
                if (!empty($value['appendNavigation'])) {
                    $appendNavigation = $value['appendNavigation'];
                }
            }
        }
        $data = $this->export($modifiedOptions, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $uuidParts = explode('-', $item['uuid']);
            if ($uuidParts[0] === 'modules') {
                if ($uuidParts[2] === 'settings') {
                    $typeName = 'settings.json';
                } else {
                    $typeName = $uuidParts[2] . '-layout.json';
                }
                $filename = $tmpDirectory . 'modules/' . $this->cleanFileName($uuidParts[1]) . '/' . $typeName;
                $metaData[$uuidParts[1]][] = $uuidParts[2];
            } else {
                if ($item['uuid'] == 'app') {
                    $item['uuid'] = 'navigation';
                    $metaData[$item['uuid']] = [];
                    $appendList = [];
                    $navigation = [];
                    foreach ($item['config']['navigation'] as $value) {
                        if (in_array($value['title'], $appendNavigation)) {
                            $appendList[] = $value;
                            array_push($navigation, $value['title']);
                        }
                    }
                    $metaData[$item['uuid']] = $navigation;
                    $item['config']['navigation'] =  $appendList;
                }
                $filename = $tmpDirectory . 'views/' . $item['uuid'] . '.json';
            }
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        if ($type == 'viewTemplates') {
            $viewMetaData = [];
            foreach ($metaData as $key => $value) {
                $mmd = $this->getMetadataByType($key);
                $viewMetaData[$key] =['name' => $mmd->getDescriptions()['plural'] , 'apiName' => $key, 'views' => $value];
            }
            $metaData = $viewMetaData;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return void
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        // Module views
        if ($key == 'viewTemplates') {
            $folderPath = $tmpDirectory . 'modules/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $folder) {
                $viewPaths = [
                    $folderPath . $folder . '/detail-layout.json',
                    $folderPath . $folder . '/list-layout.json',
                    $folderPath . $folder . '/form-layout.json',
                    $folderPath . $folder . '/settings.json'
                ];
                foreach ($viewPaths as $viewPath) {
                    if (file_exists($viewPath)) {
                        $fileData = file_get_contents($viewPath);
                        $jsonData = json_decode($fileData, true);
                        $data[] = $jsonData;
                    }
                }
            }
        } else {
            // Other views
            $folderPath = $tmpDirectory . $key . '/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $file) {
                $viewPath = $folderPath . $file;
                if (file_exists($viewPath)) {
                    $fileData = file_get_contents($viewPath);
                    $jsonData = json_decode($fileData, true);
                    if ($jsonData['uuid'] == 'navigation') {
                        $jsonData['uuid'] = 'app';
                    }
                    $data[] = $jsonData;
                }
            }
        }

        return $data;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            if ($value['uuid'] == 'app') {
                $value['appendNavigation'] = [];
                foreach ($value['config']['navigation'] as $item) {
                    $value['appendNavigation'][] = $item['title'];
                }
                array_push($exportOption, ['value' => $value['uuid'], 'appendNavigation' => $value['appendNavigation']]);
            } else {
                array_push($exportOption, $value['uuid']);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $item) {
            if (is_array($item)) {
                if ($item["value"] == "app") {
                    $content["navigation"] = $item["appendNavigation"];
                }
            } else {
                $moduleParts = explode('-', $item);
                if (!array_key_exists($moduleParts[1], $content)) {
                    $content[$moduleParts[1]] = [];
                }
                array_push($content[$moduleParts[1]], $moduleParts[2]);
            }
        }

        return $content;
    }

    public function getMetadataByType($type)
    {
        $repository = $this->entityManager->getRepository(ModelMetadata::class);
        return $repository->findOneBy(array('type' => $type));
    }

}
